"""Test equipment command execution side effects."""

from unittest.mock import sentinel
import inspect
from datetime import datetime
from decoy import Decoy, matchers
from typing import Any, Optional, cast, Dict

import pytest
from _pytest.fixtures import SubRequest

from opentrons_shared_data.pipette.types import (
    PipetteNameType,
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.labware.types import LabwareUri

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.types import Mount as HwMount, MountType, DeckSlotName, Point
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules import (
    TempDeck,
    MagDeck,
    HeaterShaker,
    AbstractModule,
)
from opentrons.hardware_control.dev_types import PipetteDict

from opentrons.protocol_engine import errors
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    DeckType,
    ModuleLocation,
    OnLabwareLocation,
    NonStackedLocation,
    LoadedPipette,
    LabwareOffset,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    OnAddressableAreaOffsetLocationSequenceComponent,
    OnModuleOffsetLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    ModuleModel,
    ModuleDefinition,
    OFF_DECK_LOCATION,
    FlowRates,
    AddressableAreaLocation,
)

from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.state.modules import HardwareModule
from opentrons.protocol_engine.resources import (
    ModelUtils,
    LabwareDataProvider,
    ModuleDataProvider,
    pipette_data_provider,
)
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.execution.equipment import (
    EquipmentHandler,
    LoadedPipetteData,
    LoadedModuleData,
    LoadedLabwareData,
)
from ..pipette_fixtures import get_default_nozzle_map
from opentrons.protocol_engine.resources import deck_configuration_provider


def _make_config(use_virtual_modules: bool) -> Config:
    return Config(
        use_virtual_modules=use_virtual_modules,
        # Robot and deck type are arbitrary.
        robot_type="OT-2 Standard",
        deck_type=DeckType.OT2_STANDARD,
    )


@pytest.fixture
def available_sensors() -> pipette_definition.AvailableSensorDefinition:
    """Provide a list of sensors."""
    return pipette_definition.AvailableSensorDefinition(
        sensors=["pressure", "capacitive", "environment"]
    )


@pytest.fixture(autouse=True)
def patch_mock_pipette_data_provider(
    decoy: Decoy,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Mock out pipette_data_provider top level functions."""
    for name, func in inspect.getmembers(pipette_data_provider, inspect.isfunction):
        monkeypatch.setattr(pipette_data_provider, name, decoy.mock(func=func))


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore instance."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mocked out ModelUtils instance."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def labware_data_provider(decoy: Decoy) -> LabwareDataProvider:
    """Get a mocked out LabwareDataProvider instance."""
    return decoy.mock(cls=LabwareDataProvider)


@pytest.fixture
def module_data_provider(decoy: Decoy) -> ModuleDataProvider:
    """Get a mocked out ModuleDataProvider instance."""
    return decoy.mock(cls=ModuleDataProvider)


@pytest.fixture
async def temp_module_v1(decoy: Decoy) -> TempDeck:
    """Get a mocked out module fixture."""
    temp_mod = decoy.mock(cls=TempDeck)
    decoy.when(temp_mod.device_info).then_return({"serial": "serial-1"})
    decoy.when(temp_mod.model()).then_return("temperatureModuleV1")

    return temp_mod


@pytest.fixture
async def temp_module_v2(decoy: Decoy) -> TempDeck:
    """Get a mocked out module fixture."""
    temp_mod = decoy.mock(cls=TempDeck)
    decoy.when(temp_mod.device_info).then_return({"serial": "serial-2"})
    decoy.when(temp_mod.model()).then_return("temperatureModuleV2")

    return temp_mod


@pytest.fixture(params=["v0", "v1", "v3"])
def tip_overlap_versions(request: SubRequest) -> str:
    """Get a series of tip overlap versions."""
    return cast(str, request.param)


@pytest.fixture
def loaded_static_pipette_data(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    target_tip_overlap_data: Dict[str, float],
    available_sensors: pipette_definition.AvailableSensorDefinition,
) -> LoadedStaticPipetteData:
    """Get a pipette config data value object."""
    return LoadedStaticPipetteData(
        model="pipette_model",
        display_name="pipette name",
        min_volume=1.23,
        max_volume=4.56,
        channels=7,
        flow_rates=FlowRates(
            default_blow_out={"a": 1.23},
            default_aspirate={"b": 4.56},
            default_dispense={"c": 7.89},
        ),
        tip_configuration_lookup_table={4.56: supported_tip_fixture},
        nominal_tip_overlap=target_tip_overlap_data,
        home_position=10.11,
        nozzle_offset_z=12.13,
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        back_left_corner_offset=Point(x=1, y=2, z=3),
        front_right_corner_offset=Point(x=4, y=5, z=6),
        pipette_lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
    )


@pytest.fixture
def target_tip_overlap_data(tip_overlap_versions: str) -> Dict[str, float]:
    """Get the corresponding overlap data for the version."""
    return {"default": 2.13 * int(tip_overlap_versions[1:])}


@pytest.fixture
def virtual_pipette_data_provider(
    decoy: Decoy,
) -> pipette_data_provider.VirtualPipetteDataProvider:
    """Virtual pipette data provider."""
    return decoy.mock(cls=pipette_data_provider.VirtualPipetteDataProvider)


@pytest.fixture
def subject(
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    labware_data_provider: LabwareDataProvider,
    module_data_provider: ModuleDataProvider,
    model_utils: ModelUtils,
    virtual_pipette_data_provider: pipette_data_provider.VirtualPipetteDataProvider,
) -> EquipmentHandler:
    """Get an EquipmentHandler test subject with its dependencies mocked out."""
    return EquipmentHandler(
        hardware_api=hardware_api,
        state_store=state_store,
        labware_data_provider=labware_data_provider,
        module_data_provider=module_data_provider,
        model_utils=model_utils,
        virtual_pipette_data_provider=virtual_pipette_data_provider,
    )


async def test_load_labware(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    labware_data_provider: LabwareDataProvider,
    subject: EquipmentHandler,
) -> None:
    """It should load labware definition and offset data and generate an ID."""
    decoy.when(model_utils.generate_id()).then_return("unique-id")

    decoy.when(state_store.labware.get_definition_by_uri(matchers.IsA(str))).then_raise(
        errors.LabwareDefinitionDoesNotExistError("oh no")
    )

    decoy.when(
        await labware_data_provider.get_labware_definition(
            load_name="load-name",
            namespace="opentrons-test",
            version=1,
        )
    ).then_return(sentinel.labware_def)
    decoy.when(
        state_store.geometry.get_projected_offset_location(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3), None
        )
    ).then_return(
        [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3")]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_3),
            locationSequence=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = await subject.load_labware(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
        labware_id=None,
    )

    assert result == LoadedLabwareData(
        labware_id="unique-id",
        definition=sentinel.labware_def,
        offsetId="labware-offset-id",
    )


async def test_load_labware_off_deck(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should load labware definition and offset data and generate an ID."""
    decoy.when(model_utils.generate_id()).then_return("unique-id")

    decoy.when(
        state_store.labware.get_definition_by_uri(
            cast("LabwareUri", "opentrons-test/load-name/1")
        )
    ).then_return(sentinel.labware_def)

    result = await subject.load_labware(
        location=OFF_DECK_LOCATION,
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
        labware_id=None,
    )

    assert result == LoadedLabwareData(
        labware_id="unique-id",
        definition=sentinel.labware_def,
        offsetId=None,
    )


async def test_load_labware_uses_provided_id(
    decoy: Decoy,
    state_store: StateStore,
    labware_data_provider: LabwareDataProvider,
    subject: EquipmentHandler,
) -> None:
    """It should use the provided ID rather than generating an ID for the labware."""
    decoy.when(state_store.labware.get_definition_by_uri(matchers.IsA(str))).then_raise(
        errors.LabwareDefinitionDoesNotExistError("oh no")
    )

    decoy.when(
        await labware_data_provider.get_labware_definition(
            load_name="load-name",
            namespace="opentrons-test",
            version=1,
        )
    ).then_return(sentinel.labware_def)
    decoy.when(
        state_store.geometry.get_projected_offset_location(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3), None
        )
    ).then_return(
        [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3")]
    )
    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
        )
    ).then_return(None)

    result = await subject.load_labware(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
        labware_id="my-labware-id",
    )

    assert result == LoadedLabwareData(
        labware_id="my-labware-id", definition=sentinel.labware_def, offsetId=None
    )


async def test_load_labware_uses_loaded_labware_def(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    labware_data_provider: LabwareDataProvider,
    subject: EquipmentHandler,
) -> None:
    """Loading labware should use the labware definition already in state."""
    expected_uri = uri_from_details(
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
    )

    decoy.when(model_utils.generate_id()).then_return("unique-id")

    decoy.when(state_store.labware.get_definition_by_uri(expected_uri)).then_return(
        sentinel.labware_def
    )

    decoy.when(
        state_store.geometry.get_projected_offset_location(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3), None
        )
    ).then_return(
        [OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3")]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
        )
    ).then_return(None)

    result = await subject.load_labware(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
        labware_id=None,
    )

    assert result == LoadedLabwareData(
        labware_id="unique-id",
        definition=sentinel.labware_def,
        offsetId=None,
    )

    decoy.verify(
        await labware_data_provider.get_labware_definition(
            load_name="load-name",
            namespace="opentrons-test",
            version=1,
        ),
        times=0,
    )


async def test_load_labware_on_module(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should load labware definition and offset data and generate an ID."""
    decoy.when(model_utils.generate_id()).then_return("unique-id")

    decoy.when(
        state_store.labware.get_definition_by_uri(matchers.IsA(str))
    ).then_return(sentinel.labware_def)

    decoy.when(state_store.modules.get_requested_model("module-id")).then_return(
        ModuleModel.THERMOCYCLER_MODULE_V1
    )
    decoy.when(state_store.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_3)
    )

    decoy.when(
        state_store.geometry.get_projected_offset_location(
            ModuleLocation(moduleId="module-id"), None
        )
    ).then_return(
        [
            OnModuleOffsetLocationSequenceComponent(
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1"),
        ]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_3,
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1,
            ),
            locationSequence=[
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = await subject.load_labware(
        location=ModuleLocation(moduleId="module-id"),
        load_name="load-name",
        namespace="opentrons-test",
        version=1,
        labware_id=None,
    )

    assert result == LoadedLabwareData(
        labware_id="unique-id",
        definition=sentinel.labware_def,
        offsetId="labware-offset-id",
    )


def test_find_offset_id_of_labware_on_deck_slot(
    decoy: Decoy,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should find the offset by resolving the provided location."""
    decoy.when(
        state_store.geometry.get_projected_offset_location(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_3), None
        )
    ).then_return(
        [
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3"),
        ]
    )
    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_3,
                moduleModel=None,
            ),
            locationSequence=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                )
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name/1",
        labware_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
    )

    assert result == "labware-offset-id"


def test_find_offset_id_of_labware_on_module(
    decoy: Decoy,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should find a new offset by resolving the new location."""
    decoy.when(state_store.modules.get_requested_model("input-module-id")).then_return(
        ModuleModel.THERMOCYCLER_MODULE_V1
    )
    decoy.when(state_store.modules.get_location("input-module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_3)
    )

    decoy.when(
        state_store.geometry.get_projected_offset_location(
            ModuleLocation(moduleId="input-module-id"), None
        )
    ).then_return(
        [
            OnModuleOffsetLocationSequenceComponent(
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3"),
        ]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name/1",
            location=[
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                ),
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_3,
                moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1,
            ),
            locationSequence=[
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.THERMOCYCLER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="3"
                ),
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name/1",
        labware_location=ModuleLocation(moduleId="input-module-id"),
    )

    assert result == "labware-offset-id"


@pytest.mark.parametrize(
    argnames=["parent_location", "expected_result"],
    argvalues=[
        (DeckSlotLocation(slotName=DeckSlotName.SLOT_1), "labware-offset-id"),
        (OFF_DECK_LOCATION, None),
    ],
)
def test_find_offset_id_of_labware_on_labware(
    decoy: Decoy,
    parent_location: NonStackedLocation,
    expected_result: Optional[str],
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should find an offset for a labware on a labware."""
    decoy.when(state_store.labware.get_definition_uri("labware-id")).then_return(
        LabwareUri("opentrons-test/load-name-2/1")
    )

    decoy.when(state_store.labware.get_parent_location("labware-id")).then_return(
        parent_location
    )
    decoy.when(
        state_store.geometry.get_projected_offset_location(
            OnLabwareLocation(labwareId="labware-id"), None
        )
    ).then_return(
        [
            OnLabwareOffsetLocationSequenceComponent(
                labwareUri="opentrons-test/load-name-2/1"
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1"),
        ]
        if parent_location is not OFF_DECK_LOCATION
        else None
    )
    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name-1/1",
            location=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                definitionUri="opentrons-test/load-name-2/1",
            ),
            locationSequence=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name-1/1",
        labware_location=OnLabwareLocation(labwareId="labware-id"),
    )

    assert result == expected_result


def test_find_offset_id_of_labware_on_labware_on_modules(
    decoy: Decoy,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should find an offset for a labware on a labware on a module."""
    decoy.when(state_store.labware.get_definition_uri("labware-id")).then_return(
        LabwareUri("opentrons-test/load-name-2/1")
    )

    decoy.when(state_store.labware.get_parent_location("labware-id")).then_return(
        ModuleLocation(moduleId="module-id"),
    )

    decoy.when(state_store.modules.get_requested_model("module-id")).then_return(
        ModuleModel.HEATER_SHAKER_MODULE_V1
    )

    decoy.when(state_store.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )

    decoy.when(
        state_store.geometry.get_projected_offset_location(
            OnLabwareLocation(labwareId="labware-id"), None
        )
    ).then_return(
        [
            OnLabwareOffsetLocationSequenceComponent(
                labwareUri="opentrons-test/load-name-2/1"
            ),
            OnModuleOffsetLocationSequenceComponent(
                moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1"),
        ]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name-1/1",
            location=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1,
                definitionUri="opentrons-test/load-name-2/1",
            ),
            locationSequence=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name-1/1",
        labware_location=OnLabwareLocation(labwareId="labware-id"),
    )

    assert result == "labware-offset-id"


def test_find_offset_id_of_labware_on_labware_on_labware_modules(
    decoy: Decoy,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """It should find an offset for a labware on a labware on a module."""
    decoy.when(state_store.labware.get_definition_uri("labware-id")).then_return(
        LabwareUri("opentrons-test/load-name-2/1")
    )

    decoy.when(state_store.labware.get_parent_location("labware-id")).then_return(
        ModuleLocation(moduleId="labware-id-2"),
    )

    decoy.when(state_store.labware.get_definition_uri("labware-id-2")).then_return(
        LabwareUri("opentrons-test/load-name-3/1")
    )

    decoy.when(state_store.labware.get_parent_location("labware-id-2")).then_return(
        ModuleLocation(moduleId="module-id"),
    )

    decoy.when(state_store.modules.get_requested_model("module-id")).then_return(
        ModuleModel.HEATER_SHAKER_MODULE_V1
    )

    decoy.when(state_store.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )

    decoy.when(
        state_store.geometry.get_projected_offset_location(
            OnLabwareLocation(labwareId="labware-id"), None
        )
    ).then_return(
        [
            OnLabwareOffsetLocationSequenceComponent(
                labwareUri="opentrons-test/load-name-2/1"
            ),
            OnLabwareOffsetLocationSequenceComponent(
                labwareUri="opentrons-test/load-name-3/1"
            ),
            OnModuleOffsetLocationSequenceComponent(
                moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1"),
        ]
    )

    decoy.when(
        state_store.labware.find_applicable_labware_offset(
            definition_uri="opentrons-test/load-name-1/1",
            location=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-3/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
        )
    ).then_return(
        LabwareOffset(
            id="labware-offset-id",
            createdAt=datetime(year=2021, month=1, day=2),
            definitionUri="opentrons-test/load-name/1",
            location=LegacyLabwareOffsetLocation(
                slotName=DeckSlotName.SLOT_1,
                moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1,
                definitionUri="opentrons-test/load-name-2/1",
            ),
            locationSequence=[
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-2/1"
                ),
                OnLabwareOffsetLocationSequenceComponent(
                    labwareUri="opentrons-test/load-name-3/1"
                ),
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.HEATER_SHAKER_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name-1/1",
        labware_location=OnLabwareLocation(labwareId="labware-id"),
    )

    assert result == "labware-offset-id"


def test_find_offset_id_of_labware_off_deck(subject: EquipmentHandler) -> None:
    """It should return None for offset_id of labware off-deck."""
    result = subject.find_applicable_labware_offset_id(
        labware_definition_uri="opentrons-test/load-name/1",
        labware_location=OFF_DECK_LOCATION,
    )
    assert result is None


async def test_load_pipette(
    decoy: Decoy,
    model_utils: ModelUtils,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    loaded_static_pipette_data: LoadedStaticPipetteData,
    tip_overlap_versions: str,
    subject: EquipmentHandler,
) -> None:
    """It should load pipette data, check attachment, and generate an ID."""
    pipette_dict = cast(PipetteDict, {"model": "hello", "pipette_id": "world"})

    decoy.when(state_store.config.use_virtual_pipettes).then_return(False)
    decoy.when(model_utils.generate_id()).then_return("unique-id")
    decoy.when(state_store.pipettes.get_by_mount(MountType.RIGHT)).then_return(
        LoadedPipette.model_construct(pipetteName=PipetteNameType.P300_MULTI)  # type: ignore[call-arg]
    )
    decoy.when(hardware_api.get_attached_instrument(mount=HwMount.LEFT)).then_return(
        pipette_dict
    )

    decoy.when(
        pipette_data_provider.validate_and_default_tip_overlap_version(
            tip_overlap_versions
        )
    ).then_return(tip_overlap_versions)
    decoy.when(
        pipette_data_provider.get_pipette_static_config(
            pipette_dict=pipette_dict, tip_overlap_version=tip_overlap_versions
        ),
    ).then_return(loaded_static_pipette_data)

    decoy.when(hardware_api.get_instrument_max_height(mount=HwMount.LEFT)).then_return(
        42.0
    )

    result = await subject.load_pipette(
        pipette_name=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
        pipette_id=None,
        tip_overlap_version=tip_overlap_versions,
    )

    assert result == LoadedPipetteData(
        pipette_id="unique-id",
        serial_number="world",
        static_config=loaded_static_pipette_data,
    )

    decoy.verify(
        await hardware_api.cache_instruments(
            {
                HwMount.LEFT: PipetteNameType.P300_SINGLE.value,
                HwMount.RIGHT: PipetteNameType.P300_MULTI.value,
            }
        ),
    )


async def test_load_pipette_96_channels(
    decoy: Decoy,
    model_utils: ModelUtils,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    loaded_static_pipette_data: LoadedStaticPipetteData,
    tip_overlap_versions: str,
    subject: EquipmentHandler,
) -> None:
    """It should load pipette data, check attachment, and generate an ID."""
    pipette_dict = cast(PipetteDict, {"model": "hello", "pipette_id": "world"})

    decoy.when(state_store.config.use_virtual_pipettes).then_return(False)
    decoy.when(model_utils.generate_id()).then_return("unique-id")
    decoy.when(hardware_api.get_attached_instrument(mount=HwMount.LEFT)).then_return(
        pipette_dict
    )
    decoy.when(
        pipette_data_provider.validate_and_default_tip_overlap_version(
            tip_overlap_versions
        )
    ).then_return(tip_overlap_versions)
    decoy.when(
        pipette_data_provider.get_pipette_static_config(
            pipette_dict=pipette_dict, tip_overlap_version=tip_overlap_versions
        )
    ).then_return(loaded_static_pipette_data)

    decoy.when(hardware_api.get_instrument_max_height(mount=HwMount.LEFT)).then_return(
        42.0
    )

    result = await subject.load_pipette(
        pipette_name=PipetteNameType.P1000_96,
        mount=MountType.LEFT,
        pipette_id=None,
        tip_overlap_version=tip_overlap_versions,
    )

    assert result == LoadedPipetteData(
        pipette_id="unique-id",
        serial_number="world",
        static_config=loaded_static_pipette_data,
    )


async def test_load_pipette_uses_provided_id(
    decoy: Decoy,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    loaded_static_pipette_data: LoadedStaticPipetteData,
    tip_overlap_versions: str,
    subject: EquipmentHandler,
) -> None:
    """It should use the provided ID rather than generating an ID for the pipette."""
    pipette_dict = cast(PipetteDict, {"model": "hello", "pipette_id": "world"})

    decoy.when(state_store.config.use_virtual_pipettes).then_return(False)
    decoy.when(hardware_api.get_attached_instrument(mount=HwMount.LEFT)).then_return(
        pipette_dict
    )
    decoy.when(
        pipette_data_provider.validate_and_default_tip_overlap_version(
            tip_overlap_versions
        )
    ).then_return(tip_overlap_versions)
    decoy.when(
        pipette_data_provider.get_pipette_static_config(
            pipette_dict=pipette_dict, tip_overlap_version=tip_overlap_versions
        )
    ).then_return(loaded_static_pipette_data)

    result = await subject.load_pipette(
        pipette_name=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
        pipette_id="my-pipette-id",
        tip_overlap_version=tip_overlap_versions,
    )

    assert result == LoadedPipetteData(
        pipette_id="my-pipette-id",
        serial_number="world",
        static_config=loaded_static_pipette_data,
    )


async def test_load_pipette_use_virtual(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    loaded_static_pipette_data: LoadedStaticPipetteData,
    subject: EquipmentHandler,
    tip_overlap_versions: str,
    virtual_pipette_data_provider: pipette_data_provider.VirtualPipetteDataProvider,
) -> None:
    """It should use the provided ID rather than generating an ID for the pipette."""
    decoy.when(state_store.config.use_virtual_pipettes).then_return(True)
    decoy.when(state_store.config.robot_type).then_return("OT-2 Standard")
    decoy.when(model_utils.generate_id()).then_return("unique-id")
    decoy.when(model_utils.generate_id(prefix="fake-serial-number-")).then_return(
        "fake-serial"
    )
    decoy.when(
        pipette_data_provider.validate_and_default_tip_overlap_version(
            tip_overlap_versions
        )
    ).then_return(tip_overlap_versions)
    decoy.when(
        virtual_pipette_data_provider.get_virtual_pipette_static_config(
            PipetteNameType.P300_SINGLE.value, "unique-id", tip_overlap_versions
        )
    ).then_return(loaded_static_pipette_data)

    result = await subject.load_pipette(
        pipette_name=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
        pipette_id=None,
        tip_overlap_version=tip_overlap_versions,
    )

    assert result == LoadedPipetteData(
        pipette_id="unique-id",
        serial_number="fake-serial",
        static_config=loaded_static_pipette_data,
    )


async def test_load_pipette_raises_if_pipette_not_attached(
    decoy: Decoy,
    model_utils: ModelUtils,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    subject: EquipmentHandler,
) -> None:
    """Loading a pipette should raise if unable to cache instruments."""
    decoy.when(state_store.config.use_virtual_pipettes).then_return(False)

    decoy.when(model_utils.generate_id()).then_return("unique-id")

    decoy.when(
        await hardware_api.cache_instruments(
            {HwMount.LEFT: PipetteNameType.P300_SINGLE.value}
        )
    ).then_raise(
        RuntimeError(
            "mount LEFT: instrument p300_single was requested, "
            "but no instrument is present"
        )
    )

    with pytest.raises(
        errors.FailedToLoadPipetteError, match=".+p300_single was requested"
    ):
        await subject.load_pipette(
            pipette_name=PipetteNameType.P300_SINGLE,
            mount=MountType.LEFT,
            pipette_id=None,
            tip_overlap_version="v9999",
        )


async def test_load_module(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    module_data_provider: ModuleDataProvider,
    hardware_api: HardwareControlAPI,
    tempdeck_v1_def: ModuleDefinition,
    tempdeck_v2_def: ModuleDefinition,
    temp_module_v1: AbstractModule,
    temp_module_v2: AbstractModule,
    subject: EquipmentHandler,
) -> None:
    """It should load a module, returning its ID, serial & definition in result."""
    decoy.when(model_utils.ensure_id("input-module-id")).then_return("module-id")

    decoy.when(
        module_data_provider.get_definition(ModuleModel.TEMPERATURE_MODULE_V1)
    ).then_return(tempdeck_v1_def)

    decoy.when(
        module_data_provider.get_definition(ModuleModel.TEMPERATURE_MODULE_V2)
    ).then_return(tempdeck_v2_def)

    decoy.when(hardware_api.attached_modules).then_return(
        [
            temp_module_v1,
            temp_module_v2,
        ]
    )

    decoy.when(state_store.config).then_return(_make_config(use_virtual_modules=False))

    decoy.when(
        state_store.geometry._addressable_areas.get_addressable_area_base_slot(
            DeckSlotName.SLOT_1.value
        )
    ).then_return(DeckSlotName.SLOT_1)
    decoy.when(
        state_store.geometry._addressable_areas.get_cutout_id_by_deck_slot_name(
            slot_name=DeckSlotName.SLOT_1
        )
    ).then_return("cutout1")

    decoy.when(
        state_store.modules.select_hardware_module_to_load(
            model=ModuleModel.TEMPERATURE_MODULE_V1,
            location=deck_configuration_provider.get_cutout_id_by_deck_slot_name(
                slot_name=DeckSlotName.SLOT_1
            ),
            attached_modules=[
                HardwareModule(serial_number="serial-1", definition=tempdeck_v1_def),
                HardwareModule(serial_number="serial-2", definition=tempdeck_v2_def),
            ],
            expected_serial_number=None,
        )
    ).then_return(HardwareModule(serial_number="serial-1", definition=tempdeck_v1_def))

    result = await subject.load_module(
        model=ModuleModel.TEMPERATURE_MODULE_V1,
        location=AddressableAreaLocation(addressableAreaName=DeckSlotName.SLOT_1.value),
        module_id="input-module-id",
    )

    assert result == LoadedModuleData(
        module_id="module-id",
        serial_number="serial-1",
        definition=tempdeck_v1_def,
    )


async def test_load_module_using_virtual(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    module_data_provider: ModuleDataProvider,
    hardware_api: HardwareControlAPI,
    tempdeck_v1_def: ModuleDefinition,
    tempdeck_v2_def: ModuleDefinition,
    temp_module_v1: AbstractModule,
    temp_module_v2: AbstractModule,
    subject: EquipmentHandler,
) -> None:
    """It should load a virtual module."""
    decoy.when(model_utils.ensure_id("input-module-id")).then_return("module-id")

    decoy.when(model_utils.generate_id(prefix="fake-serial-number-")).then_return(
        "fake-serial-number-abc123"
    )

    decoy.when(
        module_data_provider.get_definition(ModuleModel.TEMPERATURE_MODULE_V1)
    ).then_return(tempdeck_v1_def)

    decoy.when(state_store.config).then_return(_make_config(use_virtual_modules=True))

    result = await subject.load_module(
        model=ModuleModel.TEMPERATURE_MODULE_V1,
        location=AddressableAreaLocation(addressableAreaName=DeckSlotName.SLOT_1.value),
        module_id="input-module-id",
    )

    assert result == LoadedModuleData(
        module_id="module-id",
        serial_number="fake-serial-number-abc123",
        definition=tempdeck_v1_def,
    )


async def test_load_magnetic_block(
    decoy: Decoy,
    model_utils: ModelUtils,
    state_store: StateStore,
    module_data_provider: ModuleDataProvider,
    hardware_api: HardwareControlAPI,
    mag_block_v1_def: ModuleDefinition,
    subject: EquipmentHandler,
) -> None:
    """It should load a mag block, returning its ID & definition in result."""
    decoy.when(model_utils.ensure_id("input-module-id")).then_return("module-id")

    decoy.when(
        module_data_provider.get_definition(ModuleModel.MAGNETIC_BLOCK_V1)
    ).then_return(mag_block_v1_def)

    result = await subject.load_magnetic_block(
        model=ModuleModel.MAGNETIC_BLOCK_V1,
        location=AddressableAreaLocation(addressableAreaName=DeckSlotName.SLOT_1.value),
        module_id="input-module-id",
    )

    assert result == LoadedModuleData(
        module_id="module-id",
        serial_number=None,
        definition=mag_block_v1_def,
    )


def test_get_module_hardware_api(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareControlAPI,
    subject: EquipmentHandler,
) -> None:
    """It should get a module's hardware API."""
    module_1 = decoy.mock(cls=TempDeck)
    module_2 = decoy.mock(cls=MagDeck)
    module_3 = decoy.mock(cls=HeaterShaker)

    decoy.when(state_store.config).then_return(_make_config(use_virtual_modules=False))
    decoy.when(state_store.modules.get_serial_number("module-id")).then_return(
        "serial-2"
    )

    decoy.when(module_1.device_info).then_return({"serial": "serial-1"})
    decoy.when(module_2.device_info).then_return({"serial": "serial-2"})
    decoy.when(module_3.device_info).then_return({"serial": "serial-3"})
    decoy.when(hardware_api.attached_modules).then_return(
        [module_1, module_2, module_3]
    )

    result = subject.get_module_hardware_api(cast(Any, "module-id"))
    assert result is module_2


def test_get_module_hardware_api_virtual(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareControlAPI,
    subject: EquipmentHandler,
) -> None:
    """It should return None if modules are being virtualized."""
    module_1 = decoy.mock(cls=TempDeck)
    module_2 = decoy.mock(cls=MagDeck)
    module_3 = decoy.mock(cls=HeaterShaker)

    decoy.when(state_store.config).then_return(_make_config(use_virtual_modules=True))
    decoy.when(state_store.modules.get_serial_number("module-id")).then_return(
        "serial-2"
    )

    decoy.when(module_1.device_info).then_return({"serial": "serial-1"})
    decoy.when(module_2.device_info).then_return({"serial": "serial-2"})
    decoy.when(module_3.device_info).then_return({"serial": "serial-3"})
    decoy.when(hardware_api.attached_modules).then_return(
        [module_1, module_2, module_3]
    )

    result = subject.get_module_hardware_api(cast(Any, "module-id"))
    assert result is None


def test_get_module_hardware_api_missing(
    decoy: Decoy,
    state_store: StateStore,
    hardware_api: HardwareControlAPI,
    subject: EquipmentHandler,
) -> None:
    """It should raise an error if a module's hardware API is not found."""
    module_1 = decoy.mock(cls=TempDeck)
    module_2 = decoy.mock(cls=MagDeck)
    module_3 = decoy.mock(cls=HeaterShaker)

    decoy.when(state_store.config).then_return(_make_config(use_virtual_modules=False))
    decoy.when(state_store.modules.get_serial_number("module-id")).then_return(
        "the-limit-does-not-exist"
    )

    decoy.when(module_1.device_info).then_return({"serial": "serial-1"})
    decoy.when(module_2.device_info).then_return({"serial": "serial-2"})
    decoy.when(module_3.device_info).then_return({"serial": "serial-3"})
    decoy.when(hardware_api.attached_modules).then_return(
        [module_1, module_2, module_3]
    )

    with pytest.raises(errors.ModuleNotAttachedError):
        subject.get_module_hardware_api(cast(Any, "module-id"))



from __future__ import annotations

from typing_extensions import Literal, Final, TypedDict
from typing import Optional, List, Sequence, TYPE_CHECKING, Union, Tuple
from opentrons.hardware_control.modules import ThermocyclerStep

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Well
    from opentrons.protocol_api.disposal_locations import TrashBin, WasteChute
    from opentrons.protocol_api._liquid import LiquidClass
    from opentrons.protocol_api._nozzle_layout import NozzleLayout

from opentrons.types import Location, Mount, AxisMapType


# type for subscriptions
COMMAND: Final = "command"

# Robot #

DELAY: Final = "command.DELAY"
HOME: Final = "command.HOME"
PAUSE: Final = "command.PAUSE"
RESUME: Final = "command.RESUME"
COMMENT: Final = "command.COMMENT"
MOVE_LABWARE: Final = "command.MOVE_LABWARE"
CAPTURE_IMAGE: Final = "command.CAPTURE_IMAGE"

# Pipette #

ASPIRATE: Final = "command.ASPIRATE"
DISPENSE: Final = "command.DISPENSE"
DISPENSE_IN_DISPOSAL_LOCATION: Final = "command.DISPENSE_IN_DISPOSAL_LOCATION"
MIX: Final = "command.MIX"
CONSOLIDATE: Final = "command.CONSOLIDATE"
DISTRIBUTE: Final = "command.DISTRIBUTE"
TRANSFER: Final = "command.TRANSFER"
PICK_UP_TIP: Final = "command.PICK_UP_TIP"
DROP_TIP: Final = "command.DROP_TIP"
DROP_TIP_IN_DISPOSAL_LOCATION: Final = "command.DROP_TIP_IN_DISPOSAL_LOCATION"
BLOW_OUT: Final = "command.BLOW_OUT"
BLOW_OUT_IN_DISPOSAL_LOCATION: Final = "command.BLOW_OUT_IN_DISPOSAL_LOCATION"
AIR_GAP: Final = "command.AIR_GAP"
TOUCH_TIP: Final = "command.TOUCH_TIP"
RETURN_TIP: Final = "command.RETURN_TIP"
MOVE_TO: Final = "command.MOVE_TO"
MOVE_TO_DISPOSAL_LOCATION: Final = "command.MOVE_TO_DISPOSAL_LOCATION"
TRANSFER_WITH_LIQUID_CLASS: Final = "command.TRANSFER_WITH_LIQUID_CLASS"
DISTRIBUTE_WITH_LIQUID_CLASS: Final = "command.DISTRIBUTE_WITH_LIQUID_CLASS"
CONSOLIDATE_WITH_LIQUID_CLASS: Final = "command.CONSOLIDATE_WITH_LIQUID_CLASS"
SEAL: Final = "command.SEAL"
UNSEAL: Final = "command.UNSEAL"
PRESSURIZE: Final = "command.PRESSURIZE"
CONFIGURE_FOR_VOLUME: Final = "command.CONFIGURE_FOR_VOLUME"
CONFIGURE_NOZZLE_LAYOUT: Final = "command.CONFIGURE_NOZZLE_LAYOUT"


# Modules #

HEATER_SHAKER_SET_TARGET_TEMPERATURE: Final = (
    "command.HEATER_SHAKER_SET_TARGET_TEMPERATURE"
)
HEATER_SHAKER_WAIT_FOR_TEMPERATURE: Final = "command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE"
HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED: Final = (
    "command.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED"
)
HEATER_SHAKER_SET_SHAKE_SPEED: Final = "command.HEATER_SHAKER_SET_SHAKE_SPEED"
HEATER_SHAKER_OPEN_LABWARE_LATCH: Final = "command.HEATER_SHAKER_OPEN_LABWARE_LATCH"
HEATER_SHAKER_CLOSE_LABWARE_LATCH: Final = "command.HEATER_SHAKER_CLOSE_LABWARE_LATCH"
HEATER_SHAKER_DEACTIVATE_SHAKER: Final = "command.HEATER_SHAKER_DEACTIVATE_SHAKER"
HEATER_SHAKER_DEACTIVATE_HEATER: Final = "command.HEATER_SHAKER_DEACTIVATE_HEATER"

MAGDECK_CALIBRATE: Final = "command.MAGDECK_CALIBRATE"
MAGDECK_DISENGAGE: Final = "command.MAGDECK_DISENGAGE"
MAGDECK_ENGAGE: Final = "command.MAGDECK_ENGAGE"

TEMPDECK_DEACTIVATE: Final = "command.TEMPDECK_DEACTIVATE"
TEMPDECK_SET_TEMP: Final = "command.TEMPDECK_SET_TEMP"
TEMPDECK_AWAIT_TEMP: Final = "command.TEMPDECK_AWAIT_TEMP"

THERMOCYCLER_OPEN: Final = "command.THERMOCYCLER_OPEN"
THERMOCYCLER_CLOSE: Final = "command.THERMOCYCLER_CLOSE"
THERMOCYCLER_SET_BLOCK_TEMP: Final = "command.THERMOCYCLER_SET_BLOCK_TEMP"
THERMOCYCLER_START_SET_BLOCK_TEMP: Final = "command.THERMOCYCLER_START_SET_BLOCK_TEMP"
THERMOCYCLER_EXECUTE_PROFILE: Final = "command.THERMOCYCLER_EXECUTE_PROFILE"
THERMOCYCLER_START_EXECUTE_PROFILE: Final = "command.THERMOCYCLER_START_EXECUTE_PROFILE"
THERMOCYCLER_DEACTIVATE: Final = "command.THERMOCYCLER_DEACTIVATE"
THERMOCYCLER_WAIT_FOR_HOLD: Final = "command.THERMOCYCLER_WAIT_FOR_HOLD"
THERMOCYCLER_WAIT_FOR_TEMP: Final = "command.THERMOCYCLER_WAIT_FOR_TEMP"
THERMOCYCLER_WAIT_FOR_LID_TEMP: Final = "command.THERMOCYCLER_WAIT_FOR_LID_TEMP"
THERMOCYCLER_SET_LID_TEMP: Final = "command.THERMOCYCLER_SET_LID_TEMP"
THERMOCYCLER_START_SET_LID_TEMP: Final = "command.THERMOCYCLER_START_SET_LID_TEMP"
THERMOCYCLER_DEACTIVATE_LID: Final = "command.THERMOCYCLER_DEACTIVATE_LID"
THERMOCYCLER_DEACTIVATE_BLOCK: Final = "command.THERMOCYCLER_DEACTIVATE_BLOCK"

FLEX_STACKER_SET_STORED_LABWARE: Final = "command.FLEX_STACKER_SET_STORED_LABWARE"
FLEX_STACKER_RETRIEVE: Final = "command.FLEX_STACKER_RETRIEVE"
FLEX_STACKER_STORE: Final = "command.FLEX_STACKER_STORE"
FLEX_STACKER_EMPTY: Final = "command.FLEX_STACKER_EMPTY"
FLEX_STACKER_FILL: Final = "command.FLEX_STACKER_FILL"

# Robot #
ROBOT_MOVE_TO: Final = "command.ROBOT_MOVE_TO"
ROBOT_MOVE_AXES_TO: Final = "command.ROBOT_MOVE_AXES_TO"
ROBOT_MOVE_RELATIVE_TO: Final = "command.ROBOT_MOVE_RELATIVE_TO"
ROBOT_OPEN_GRIPPER_JAW: Final = "command.ROBOT_OPEN_GRIPPER_JAW"
ROBOT_CLOSE_GRIPPER_JAW: Final = "command.ROBOT_CLOSE_GRIPPER_JAW"

# Tasks #
WAIT_FOR_TASKS: Final = "command.WAIT_FOR_TASKS"
CREATE_TIMER: Final = "command.CREATE_TIMER"


class TextOnlyPayload(TypedDict):
    text: str


class MultiLocationPayload(TypedDict):
    locations: Sequence[Union[Location, Well]]


class OptionalMultiLocationPayload(TypedDict):
    locations: Optional[Sequence[Union[Location, Well]]]


class SingleInstrumentPayload(TypedDict):
    instrument: InstrumentContext


class MultiInstrumentPayload(TypedDict):
    instruments: Sequence[InstrumentContext]


class CommentCommandPayload(TextOnlyPayload):
    pass


class CommentCommand(TypedDict):
    name: Literal["command.COMMENT"]
    payload: CommentCommandPayload


class DelayCommandPayload(TextOnlyPayload):
    minutes: float
    seconds: float


class DelayCommand(TypedDict):
    name: Literal["command.DELAY"]
    payload: DelayCommandPayload


class PauseCommandPayload(TextOnlyPayload):
    userMessage: Optional[str]


class PauseCommand(TypedDict):
    name: Literal["command.PAUSE"]
    payload: PauseCommandPayload


class ResumeCommandPayload(TextOnlyPayload):
    pass


class ResumeCommand(TypedDict):
    name: Literal["command.RESUME"]
    payload: ResumeCommandPayload


# Module commands


class HeaterShakerSetTargetTemperaturePayload(TextOnlyPayload):
    pass


class HeaterShakerSetTargetTemperatureCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_SET_TARGET_TEMPERATURE"]
    payload: HeaterShakerSetTargetTemperaturePayload


class HeaterShakerWaitForTemperaturePayload(TextOnlyPayload):
    pass


class HeaterShakerWaitForTemperatureCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE"]
    payload: HeaterShakerWaitForTemperaturePayload


class HeaterShakerSetAndWaitForShakeSpeedPayload(TextOnlyPayload):
    pass


class HeaterShakerSetAndWaitForShakeSpeedCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED"]
    payload: HeaterShakerSetAndWaitForShakeSpeedPayload


class HeaterShakerSetShakeSpeedPayload(TextOnlyPayload):
    pass


class HeaterShakerSetShakeSpeedCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_SET_SHAKE_SPEED"]
    payload: HeaterShakerSetShakeSpeedPayload


class HeaterShakerOpenLabwareLatchPayload(TextOnlyPayload):
    pass


class HeaterShakerOpenLabwareLatchCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_OPEN_LABWARE_LATCH"]
    payload: HeaterShakerOpenLabwareLatchPayload


class HeaterShakerCloseLabwareLatchPayload(TextOnlyPayload):
    pass


class HeaterShakerCloseLabwareLatchCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_CLOSE_LABWARE_LATCH"]
    payload: HeaterShakerCloseLabwareLatchPayload


class HeaterShakerDeactivateShakerPayload(TextOnlyPayload):
    pass


class HeaterShakerDeactivateShakerCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_DEACTIVATE_SHAKER"]
    payload: HeaterShakerDeactivateShakerPayload


class HeaterShakerDeactivateHeaterPayload(TextOnlyPayload):
    pass


class HeaterShakerDeactivateHeaterCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_DEACTIVATE_HEATER"]
    payload: HeaterShakerDeactivateHeaterPayload


class MagdeckEngageCommandPayload(TextOnlyPayload):
    pass


class MagdeckEngageCommand(TypedDict):
    name: Literal["command.MAGDECK_ENGAGE"]
    payload: MagdeckEngageCommandPayload


class MagdeckDisengageCommandPayload(TextOnlyPayload):
    pass


class MagdeckDisengageCommand(TypedDict):
    name: Literal["command.MAGDECK_DISENGAGE"]
    payload: MagdeckDisengageCommandPayload


class MagdeckCalibrateCommandPayload(TextOnlyPayload):
    pass


class MagdeckCalibrateCommand(TypedDict):
    name: Literal["command.MAGDECK_CALIBRATE"]
    payload: MagdeckCalibrateCommandPayload


class TempdeckSetTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckSetTempCommand(TypedDict):
    name: Literal["command.TEMPDECK_SET_TEMP"]
    payload: TempdeckSetTempCommandPayload


class TempdeckAwaitTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckAwaitTempCommand(TypedDict):
    name: Literal["command.TEMPDECK_AWAIT_TEMP"]
    payload: TempdeckAwaitTempCommandPayload


class TempdeckDeactivateCommandPayload(TextOnlyPayload):
    pass


class TempdeckDeactivateCommand(TypedDict):
    name: Literal["command.TEMPDECK_DEACTIVATE"]
    payload: TempdeckDeactivateCommandPayload


class ThermocyclerOpenCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerOpenCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_OPEN"]
    payload: ThermocyclerOpenCommandPayload


class ThermocyclerSetBlockTempCommandPayload(TextOnlyPayload):
    temperature: float
    hold_time: Optional[float]


class ThermocyclerSetBlockTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_SET_BLOCK_TEMP"]
    payload: ThermocyclerSetBlockTempCommandPayload


class ThermocyclerStartSetBlockTempCommandPayload(TextOnlyPayload):
    temperature: float


class ThermocyclerStartSetBlockTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_START_SET_BLOCK_TEMP"]
    payload: ThermocyclerStartSetBlockTempCommandPayload


class ThermocyclerExecuteProfileCommandPayload(TextOnlyPayload):
    steps: List[ThermocyclerStep]


class ThermocyclerExecuteProfileCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_EXECUTE_PROFILE"]
    payload: ThermocyclerExecuteProfileCommandPayload


class ThermocyclerStartExecuteProfileCommandPayload(TextOnlyPayload):
    steps: List[ThermocyclerStep]


class ThermocyclerStartExecuteProfileCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_START_EXECUTE_PROFILE"]
    payload: ThermocyclerStartExecuteProfileCommandPayload


class ThermocyclerWaitForHoldCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForHoldCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_HOLD"]
    payload: ThermocyclerWaitForHoldCommandPayload


class ThermocyclerWaitForTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_TEMP"]
    payload: ThermocyclerWaitForTempCommandPayload


class ThermocyclerSetLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerSetLidTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_SET_LID_TEMP"]
    payload: ThermocyclerSetLidTempCommandPayload


class ThermocyclerStartSetLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerStartSetLidTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_START_SET_LID_TEMP"]
    payload: ThermocyclerStartSetLidTempCommandPayload


class ThermocyclerDeactivateLidCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateLidCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE_LID"]
    payload: ThermocyclerDeactivateLidCommandPayload


class ThermocyclerDeactivateBlockCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateBlockCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE_BLOCK"]
    payload: ThermocyclerDeactivateBlockCommandPayload


class ThermocyclerDeactivateCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE"]
    payload: ThermocyclerDeactivateCommandPayload


class ThermocyclerWaitForLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForLidTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_LID_TEMP"]
    payload: ThermocyclerWaitForLidTempCommandPayload


class ThermocyclerCloseCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerCloseCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_CLOSE"]
    payload: ThermocyclerCloseCommandPayload


class FlexStackerSetStoredLabwareCommand(TypedDict):
    name: Literal["command.FLEX_STACKER_SET_STORED_LABWARE"]
    payload: TextOnlyPayload


class FlexStackerRetrieveCommand(TypedDict):
    name: Literal["command.FLEX_STACKER_RETRIEVE"]
    payload: TextOnlyPayload


class FlexStackerStoreCommand(TypedDict):
    name: Literal["command.FLEX_STACKER_STORE"]
    payload: TextOnlyPayload


class FlexStackerEmptyCommand(TypedDict):
    name: Literal["command.FLEX_STACKER_EMPTY"]
    payload: TextOnlyPayload


class FlexStackerFillCommand(TypedDict):
    name: Literal["command.FLEX_STACKER_FILL"]
    payload: TextOnlyPayload


# Module command end


class HomeCommandPayload(TextOnlyPayload):
    axis: str


class HomeCommand(TypedDict):
    name: Literal["command.HOME"]
    payload: HomeCommandPayload


class AspirateDispenseCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Location
    volume: float
    rate: float
    end_location: Optional[Location]


class AspirateCommand(TypedDict):
    name: Literal["command.ASPIRATE"]
    payload: AspirateDispenseCommandPayload


class DispenseCommand(TypedDict):
    name: Literal["command.DISPENSE"]
    payload: AspirateDispenseCommandPayload


class DispenseInDisposalLocationCommandPayload(
    TextOnlyPayload, SingleInstrumentPayload
):
    location: Union[TrashBin, WasteChute]
    volume: float
    rate: float


class DispenseInDisposalLocationCommand(TypedDict):
    name: Literal["command.DISPENSE_IN_DISPOSAL_LOCATION"]
    payload: DispenseInDisposalLocationCommandPayload


class ConsolidateCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: Union[Location, Well]


class ConsolidateCommand(TypedDict):
    name: Literal["command.CONSOLIDATE"]
    payload: ConsolidateCommandPayload


class DistributeCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: Union[Location, Well]
    dest: List[Union[Location, Well]]


class DistributeCommand(TypedDict):
    name: Literal["command.DISTRIBUTE"]
    payload: DistributeCommandPayload


class TransferCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: List[Union[Location, Well]]


class TransferCommand(TypedDict):
    name: Literal["command.TRANSFER"]
    payload: TransferCommandPayload


class MixCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[None, Location, Well]
    volume: float
    repetitions: int


class MixCommand(TypedDict):
    name: Literal["command.MIX"]
    payload: MixCommandPayload


class DynamicMixCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    aspirate_start_location: Location
    dispense_start_location: Location
    aspirate_end_location: Union[None, Location]
    dispense_end_location: Union[None, Location]
    volume: float
    repetitions: int
    movement_delay: float


class DynamicMixCommand(TypedDict):
    name: Literal["command.MIX"]
    payload: DynamicMixCommandPayload


class BlowOutCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Optional[Location]


class BlowOutCommand(TypedDict):
    name: Literal["command.BLOW_OUT"]
    payload: BlowOutCommandPayload


class BlowOutInDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class BlowOutInDisposalLocationCommand(TypedDict):
    name: Literal["command.BLOW_OUT_IN_DISPOSAL_LOCATION"]
    payload: BlowOutInDisposalLocationCommandPayload


class TouchTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    pass


class TouchTipCommand(TypedDict):
    name: Literal["command.TOUCH_TIP"]
    payload: TouchTipCommandPayload


class AirGapCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    volume: Union[float, None]
    height: Union[float, None]


class AirGapCommand(TypedDict):
    name: Literal["command.AIR_GAP"]
    payload: AirGapCommandPayload


class ReturnTipCommandPayload(TextOnlyPayload):
    pass


class ReturnTipCommand(TypedDict):
    name: Literal["command.RETURN_TIP"]
    payload: ReturnTipCommandPayload


class PickUpTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Well


class PickUpTipCommand(TypedDict):
    name: Literal["command.PICK_UP_TIP"]
    payload: PickUpTipCommandPayload


class DropTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Well


class DropTipCommand(TypedDict):
    name: Literal["command.DROP_TIP"]
    payload: DropTipCommandPayload


class DropTipInDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class DropTipInDisposalLocationCommand(TypedDict):
    name: Literal["command.DROP_TIP_IN_DISPOSAL_LOCATION"]
    payload: DropTipInDisposalLocationCommandPayload


class MoveToCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Location


class MoveToCommand(TypedDict):
    name: Literal["command.MOVE_TO"]
    payload: MoveToCommandPayload


class MoveToDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class MoveToDisposalLocationCommand(TypedDict):
    name: Literal["command.MOVE_TO_DISPOSAL_LOCATION"]
    payload: MoveToDisposalLocationCommandPayload


class MoveLabwareCommandPayload(TextOnlyPayload):
    pass


class CaptureImageCommandPayload(TextOnlyPayload):
    resolution: Optional[Tuple[int, int]]
    zoom: Optional[float]
    contrast: Optional[float]
    brightness: Optional[float]
    saturation: Optional[float]


class LiquidClassCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    liquid_class: LiquidClass
    volume: float
    source: Union[Well, Sequence[Well], Sequence[Sequence[Well]]]
    destination: Union[
        Well, Sequence[Well], Sequence[Sequence[Well]], TrashBin, WasteChute
    ]


class TransferWithLiquidClassCommand(TypedDict):
    name: Literal["command.TRANSFER_WITH_LIQUID_CLASS"]
    payload: LiquidClassCommandPayload


class DistributeWithLiquidClassCommand(TypedDict):
    name: Literal["command.DISTRIBUTE_WITH_LIQUID_CLASS"]
    payload: LiquidClassCommandPayload


class ConsolidateWithLiquidClassCommand(TypedDict):
    name: Literal["command.CONSOLIDATE_WITH_LIQUID_CLASS"]
    payload: LiquidClassCommandPayload


class SealCommandPayload(TextOnlyPayload):
    instrument: InstrumentContext
    location: Union[None, Location, Well]


class UnsealCommandPayload(TextOnlyPayload):
    instrument: InstrumentContext
    location: Union[None, Location, Well]


class PressurizeCommandPayload(TextOnlyPayload):
    instrument: InstrumentContext


class ConfigureForVolumePayload(TypedDict, TextOnlyPayload):
    instrument: InstrumentContext
    volume: float


class ConfigureNozzleLayoutPayload(TypedDict, TextOnlyPayload):
    instrument: InstrumentContext
    style: NozzleLayout
    start: Union[str, None]
    end: Union[str, None]


class MoveLabwareCommand(TypedDict):
    name: Literal["command.MOVE_LABWARE"]
    payload: MoveLabwareCommandPayload


class CaptureImageCommand(TypedDict):
    name: Literal["command.CAPTURE_IMAGE"]
    payload: CaptureImageCommandPayload


class SealCommand(TypedDict):
    name: Literal["command.SEAL"]
    payload: SealCommandPayload


class UnsealCommand(TypedDict):
    name: Literal["command.UNSEAL"]
    payload: UnsealCommandPayload


class PressurizeCommand(TypedDict):
    name: Literal["command.PRESSURIZE"]
    payload: PressurizeCommandPayload


class ConfigureForVolumeCommand(TypedDict):
    name: Literal["command.CONFIGURE_FOR_VOLUME"]
    payload: ConfigureForVolumePayload


class ConfigureNozzleLayoutCommand(TypedDict):
    name: Literal["command.CONFIGURE_NOZZLE_LAYOUT"]
    payload: ConfigureNozzleLayoutPayload


# Robot Commands and Payloads
class GripperCommandPayload(TextOnlyPayload):
    pass


class RobotMoveToCommandPayload(TextOnlyPayload):
    location: Location
    mount: Mount


class RobotMoveAxisToCommandPayload(TextOnlyPayload):
    absolute_axes: AxisMapType


class RobotMoveAxisRelativeCommandPayload(TextOnlyPayload):
    relative_axes: AxisMapType


class RobotMoveToCommand(TypedDict):
    name: Literal["command.ROBOT_MOVE_TO"]
    payload: RobotMoveToCommandPayload


class RobotMoveAxisToCommand(TypedDict):
    name: Literal["command.ROBOT_MOVE_AXES_TO"]
    payload: RobotMoveAxisToCommandPayload


class RobotMoveAxisRelativeCommand(TypedDict):
    name: Literal["command.ROBOT_MOVE_RELATIVE_TO"]
    payload: RobotMoveAxisRelativeCommandPayload


class RobotOpenGripperJawCommand(TypedDict):
    name: Literal["command.ROBOT_OPEN_GRIPPER_JAW"]
    payload: GripperCommandPayload


class RobotCloseGripperJawCommand(TypedDict):
    name: Literal["command.ROBOT_CLOSE_GRIPPER_JAW"]
    payload: GripperCommandPayload


# Task Commands and Payloads


class WaitForTasksPayload(TextOnlyPayload):
    pass


class CreateTimerPayload(TextOnlyPayload):
    time: float


class WaitForTasksCommand(TypedDict):
    name: Literal["command.WAIT_FOR_TASKS"]
    payload: WaitForTasksPayload


class CreateTimerCommand(TypedDict):
    name: Literal["command.CREATE_TIMER"]
    payload: CreateTimerPayload


Command = Union[
    DropTipCommand,
    DropTipInDisposalLocationCommand,
    PickUpTipCommand,
    ReturnTipCommand,
    AirGapCommand,
    TouchTipCommand,
    BlowOutCommand,
    BlowOutInDisposalLocationCommand,
    MixCommand,
    DynamicMixCommand,
    TransferCommand,
    DistributeCommand,
    ConsolidateCommand,
    DispenseCommand,
    DispenseInDisposalLocationCommand,
    AspirateCommand,
    HomeCommand,
    HeaterShakerSetTargetTemperatureCommand,
    HeaterShakerWaitForTemperatureCommand,
    HeaterShakerSetAndWaitForShakeSpeedCommand,
    HeaterShakerSetShakeSpeedCommand,
    HeaterShakerOpenLabwareLatchCommand,
    HeaterShakerCloseLabwareLatchCommand,
    HeaterShakerDeactivateShakerCommand,
    HeaterShakerDeactivateHeaterCommand,
    ThermocyclerCloseCommand,
    ThermocyclerWaitForLidTempCommand,
    ThermocyclerDeactivateCommand,
    ThermocyclerDeactivateBlockCommand,
    ThermocyclerDeactivateLidCommand,
    ThermocyclerSetLidTempCommand,
    ThermocyclerStartSetLidTempCommand,
    ThermocyclerWaitForTempCommand,
    ThermocyclerWaitForHoldCommand,
    ThermocyclerExecuteProfileCommand,
    ThermocyclerStartExecuteProfileCommand,
    ThermocyclerSetBlockTempCommand,
    ThermocyclerStartSetBlockTempCommand,
    ThermocyclerOpenCommand,
    TempdeckDeactivateCommand,
    TempdeckAwaitTempCommand,
    TempdeckSetTempCommand,
    MagdeckCalibrateCommand,
    MagdeckDisengageCommand,
    MagdeckEngageCommand,
    ResumeCommand,
    PauseCommand,
    DelayCommand,
    CommentCommand,
    MoveToCommand,
    MoveToDisposalLocationCommand,
    MoveLabwareCommand,
    TransferWithLiquidClassCommand,
    DistributeWithLiquidClassCommand,
    ConsolidateWithLiquidClassCommand,
    SealCommand,
    UnsealCommand,
    PressurizeCommand,
    ConfigureForVolumeCommand,
    ConfigureNozzleLayoutCommand,
    CaptureImageCommand,
    # Robot commands
    RobotMoveToCommand,
    RobotMoveAxisToCommand,
    RobotMoveAxisRelativeCommand,
    RobotOpenGripperJawCommand,
    RobotCloseGripperJawCommand,
    # Flex Stacker commands
    FlexStackerSetStoredLabwareCommand,
    FlexStackerRetrieveCommand,
    FlexStackerStoreCommand,
    FlexStackerEmptyCommand,
    FlexStackerFillCommand,
    # Task commands
    WaitForTasksCommand,
    CreateTimerCommand,
]


CommandPayload = Union[
    CommentCommandPayload,
    ResumeCommandPayload,
    HeaterShakerSetTargetTemperaturePayload,
    HeaterShakerWaitForTemperaturePayload,
    HeaterShakerSetAndWaitForShakeSpeedPayload,
    HeaterShakerSetShakeSpeedPayload,
    HeaterShakerOpenLabwareLatchPayload,
    HeaterShakerCloseLabwareLatchPayload,
    HeaterShakerDeactivateShakerPayload,
    HeaterShakerDeactivateHeaterPayload,
    MagdeckEngageCommandPayload,
    MagdeckDisengageCommandPayload,
    MagdeckCalibrateCommandPayload,
    ThermocyclerOpenCommandPayload,
    ThermocyclerWaitForHoldCommandPayload,
    ThermocyclerWaitForTempCommandPayload,
    ThermocyclerSetLidTempCommandPayload,
    ThermocyclerStartSetLidTempCommandPayload,
    ThermocyclerDeactivateLidCommandPayload,
    ThermocyclerDeactivateBlockCommandPayload,
    ThermocyclerDeactivateCommandPayload,
    ThermocyclerWaitForLidTempCommand,
    ThermocyclerCloseCommandPayload,
    AirGapCommandPayload,
    ReturnTipCommandPayload,
    DropTipCommandPayload,
    DropTipInDisposalLocationCommandPayload,
    PickUpTipCommandPayload,
    TouchTipCommandPayload,
    BlowOutCommandPayload,
    BlowOutInDisposalLocationCommandPayload,
    MixCommandPayload,
    DynamicMixCommandPayload,
    TransferCommandPayload,
    DistributeCommandPayload,
    ConsolidateCommandPayload,
    AspirateDispenseCommandPayload,
    DispenseInDisposalLocationCommandPayload,
    HomeCommandPayload,
    ThermocyclerExecuteProfileCommandPayload,
    ThermocyclerStartExecuteProfileCommandPayload,
    ThermocyclerSetBlockTempCommandPayload,
    ThermocyclerStartSetBlockTempCommandPayload,
    TempdeckAwaitTempCommandPayload,
    TempdeckSetTempCommandPayload,
    PauseCommandPayload,
    DelayCommandPayload,
    MoveToCommandPayload,
    MoveToDisposalLocationCommandPayload,
    MoveLabwareCommandPayload,
    LiquidClassCommandPayload,
    SealCommandPayload,
    UnsealCommandPayload,
    PressurizeCommandPayload,
    ConfigureForVolumePayload,
    ConfigureNozzleLayoutPayload,
    CaptureImageCommandPayload,
    # Robot payloads
    RobotMoveToCommandPayload,
    RobotMoveAxisRelativeCommandPayload,
    RobotMoveAxisToCommandPayload,
    GripperCommandPayload,
    # Task payloads
    WaitForTasksPayload,
    CreateTimerPayload,
]


MessageSequenceId = Union[Literal["before"], Literal["after"]]


CommandMessageFields = TypedDict(
    "CommandMessageFields",
    {"$": MessageSequenceId, "id": str, "error": Optional[Exception]},
)


class MoveToMessage(CommandMessageFields, MoveToCommand):
    pass


class MoveToDisposalLocationMessage(
    CommandMessageFields, MoveToDisposalLocationCommand
):
    pass


class DropTipMessage(CommandMessageFields, DropTipCommand):
    pass


class DropTipInDisposalLocationMessage(
    CommandMessageFields, DropTipInDisposalLocationCommand
):
    pass


class PickUpTipMessage(CommandMessageFields, PickUpTipCommand):
    pass


class ReturnTipMessage(CommandMessageFields, ReturnTipCommand):
    pass


class AirGapMessage(CommandMessageFields, AirGapCommand):
    pass


class TouchTipMessage(CommandMessageFields, TouchTipCommand):
    pass


class BlowOutMessage(CommandMessageFields, BlowOutCommand):
    pass


class BlowOutInDisposalLocationMessage(
    CommandMessageFields, BlowOutInDisposalLocationCommand
):
    pass


class MixMessage(CommandMessageFields, MixCommand):
    pass


class DynamicMixMessage(CommandMessageFields, DynamicMixCommand):
    pass


class TransferMessage(CommandMessageFields, TransferCommand):
    pass


class DistributeMessage(CommandMessageFields, DistributeCommand):
    pass


class ConsolidateMessage(CommandMessageFields, ConsolidateCommand):
    pass


class DispenseMessage(CommandMessageFields, DispenseCommand):
    pass


class DispenseInDisposalLocationMessage(
    CommandMessageFields, DispenseInDisposalLocationCommand
):
    pass


class AspirateMessage(CommandMessageFields, AspirateCommand):
    pass


class HomeMessage(CommandMessageFields, HomeCommand):
    pass


class HeaterShakerSetTargetTemperatureMessage(
    CommandMessageFields, HeaterShakerSetTargetTemperatureCommand
):
    pass


class HeaterShakerWaitForTemperatureMessage(
    CommandMessageFields, HeaterShakerWaitForTemperatureCommand
):
    pass


class HeaterShakerSetAndWaitForShakeSpeedMessage(
    CommandMessageFields, HeaterShakerSetAndWaitForShakeSpeedCommand
):
    pass


class HeaterShakerSetShakeSpeedMessage(
    CommandMessageFields, HeaterShakerSetShakeSpeedCommand
):
    pass


class HeaterShakerOpenLabwareLatchMessage(
    CommandMessageFields, HeaterShakerOpenLabwareLatchCommand
):
    pass


class HeaterShakerCloseLabwareLatchMessage(
    CommandMessageFields, HeaterShakerCloseLabwareLatchCommand
):
    pass


class HeaterShakerDeactivateShakerMessage(
    CommandMessageFields, HeaterShakerDeactivateShakerCommand
):
    pass


class HeaterShakerDeactivateHeaterMessage(
    CommandMessageFields, HeaterShakerDeactivateHeaterCommand
):
    pass


class ThermocyclerCloseMessage(CommandMessageFields, ThermocyclerCloseCommand):
    pass


class ThermocyclerWaitForLidTempMessage(
    CommandMessageFields, ThermocyclerWaitForLidTempCommand
):
    pass


class ThermocyclerDeactivateMessage(
    CommandMessageFields, ThermocyclerDeactivateCommand
):
    pass


class ThermocyclerDeactivateBlockMessage(
    CommandMessageFields, ThermocyclerDeactivateBlockCommand
):
    pass


class ThermocyclerDeactivateLidMessage(
    CommandMessageFields, ThermocyclerDeactivateLidCommand
):
    pass


class ThermocyclerSetLidTempMessage(
    CommandMessageFields, ThermocyclerSetLidTempCommand
):
    pass


class ThermocyclerStartSetLidTempMessage(
    CommandMessageFields, ThermocyclerStartSetLidTempCommand
):
    pass


class ThermocyclerStartSetBlockTempMessage(
    CommandMessageFields, ThermocyclerStartSetBlockTempCommand
):
    pass


class ThermocyclerWaitForTempMessage(
    CommandMessageFields, ThermocyclerWaitForTempCommand
):
    pass


class ThermocyclerWaitForHoldMessage(
    CommandMessageFields, ThermocyclerWaitForHoldCommand
):
    pass


class ThermocyclerExecuteProfileMessage(
    CommandMessageFields, ThermocyclerExecuteProfileCommand
):
    pass


class ThermocyclerStartExecuteProfileMessage(
    CommandMessageFields, ThermocyclerStartExecuteProfileCommand
):
    pass


class ThermocyclerSetBlockTempMessage(
    CommandMessageFields, ThermocyclerSetBlockTempCommand
):
    pass


class ThermocyclerOpenMessage(CommandMessageFields, ThermocyclerOpenCommand):
    pass


class TempdeckDeactivateMessage(CommandMessageFields, TempdeckDeactivateCommand):
    pass


class TempdeckAwaitTempMessage(CommandMessageFields, TempdeckAwaitTempCommand):
    pass


class TempdeckSetTempMessage(CommandMessageFields, TempdeckSetTempCommand):
    pass


class MagdeckCalibrateMessage(CommandMessageFields, MagdeckCalibrateCommand):
    pass


class MagdeckDisengageMessage(CommandMessageFields, MagdeckDisengageCommand):
    pass


class MagdeckEngageMessage(CommandMessageFields, MagdeckEngageCommand):
    pass


class FlexStackerSetStoredLabwareMessage(
    CommandMessageFields, FlexStackerSetStoredLabwareCommand
):
    pass


class FlexStackerRetrieveMessage(CommandMessageFields, FlexStackerRetrieveCommand):
    pass


class FlexStackerStoreMessage(CommandMessageFields, FlexStackerStoreCommand):
    pass


class FlexStackerEmptyMessage(CommandMessageFields, FlexStackerEmptyCommand):
    pass


class FlexStackerFillMessage(CommandMessageFields, FlexStackerFillCommand):
    pass


class ResumeMessage(CommandMessageFields, ResumeCommand):
    pass


class PauseMessage(CommandMessageFields, PauseCommand):
    pass


class DelayMessage(CommandMessageFields, DelayCommand):
    pass


class CommentMessage(CommandMessageFields, CommentCommand):
    pass


class MoveLabwareMessage(CommandMessageFields, MoveLabwareCommand):
    pass


class RobotMoveToMessage(CommandMessageFields, RobotMoveToCommand):
    pass


class RobotMoveAxisToMessage(CommandMessageFields, RobotMoveAxisToCommand):
    pass


class RobotMoveAxisRelativeMessage(CommandMessageFields, RobotMoveAxisRelativeCommand):
    pass


class RobotOpenGripperJawMessage(CommandMessageFields, RobotOpenGripperJawCommand):
    pass


class RobotCloseGripperJawMessage(CommandMessageFields, RobotCloseGripperJawCommand):
    pass


class WaitForTasksMessage(CommandMessageFields, WaitForTasksCommand):
    pass


class CreateTimerMessage(CommandMessageFields, CreateTimerCommand):
    pass


CommandMessage = Union[
    DropTipMessage,
    DropTipInDisposalLocationMessage,
    PickUpTipMessage,
    ReturnTipMessage,
    AirGapMessage,
    TouchTipMessage,
    BlowOutMessage,
    BlowOutInDisposalLocationMessage,
    MixMessage,
    TransferMessage,
    DistributeMessage,
    ConsolidateMessage,
    DispenseMessage,
    DispenseInDisposalLocationMessage,
    AspirateMessage,
    HomeMessage,
    HeaterShakerSetTargetTemperatureMessage,
    HeaterShakerWaitForTemperatureMessage,
    HeaterShakerSetAndWaitForShakeSpeedMessage,
    HeaterShakerSetShakeSpeedMessage,
    HeaterShakerOpenLabwareLatchMessage,
    HeaterShakerCloseLabwareLatchMessage,
    HeaterShakerDeactivateShakerMessage,
    HeaterShakerDeactivateHeaterMessage,
    ThermocyclerCloseMessage,
    ThermocyclerWaitForLidTempMessage,
    ThermocyclerDeactivateMessage,
    ThermocyclerDeactivateBlockMessage,
    ThermocyclerDeactivateLidMessage,
    ThermocyclerSetLidTempMessage,
    ThermocyclerWaitForTempMessage,
    ThermocyclerWaitForHoldMessage,
    ThermocyclerExecuteProfileMessage,
    ThermocyclerStartExecuteProfileMessage,
    ThermocyclerSetBlockTempMessage,
    ThermocyclerOpenMessage,
    TempdeckSetTempMessage,
    TempdeckDeactivateMessage,
    MagdeckEngageMessage,
    MagdeckDisengageMessage,
    MagdeckCalibrateMessage,
    CommentMessage,
    DelayMessage,
    PauseMessage,
    ResumeMessage,
    MoveToMessage,
    MoveToDisposalLocationMessage,
    MoveLabwareMessage,
    # Robot Messages
    RobotMoveToMessage,
    RobotMoveAxisToMessage,
    RobotMoveAxisRelativeMessage,
    RobotOpenGripperJawMessage,
    RobotCloseGripperJawMessage,
    # Flex Stacker Messages
    FlexStackerSetStoredLabwareMessage,
    FlexStackerRetrieveMessage,
    FlexStackerStoreMessage,
    FlexStackerEmptyMessage,
    FlexStackerFillMessage,
    # Task Messages
    WaitForTasksMessage,
    CreateTimerMessage,
]
